import logging

from ha_services.exceptions import InvalidStateValue
from ha_services.mqtt4homeassistant.components import BaseComponent
from ha_services.mqtt4homeassistant.data_classes import NO_STATE, ComponentConfig, ComponentState
from ha_services.mqtt4homeassistant.device import MqttDevice


logger = logging.getLogger(__name__)


class BinarySensor(BaseComponent):
    """
    MQTT Switch component for Home Assistant.
    https://www.home-assistant.io/integrations/binary_sensor.mqtt/

    """

    ON = 'ON'
    OFF = 'OFF'

    def __init__(
        self,
        *,
        device: MqttDevice,
        name: str,
        uid: str,
        component: str = 'binary_sensor',
        initial_state=NO_STATE,  # set_state() must be called to set the value
        device_class: str | None = None,  # https://www.home-assistant.io/integrations/binary_sensor/#device-class
    ):
        super().__init__(
            device=device,
            name=name,
            uid=uid,
            component=component,
            initial_state=initial_state,
        )

        self.device_class = device_class

        self.state2bool = {
            self.ON: True,
            self.OFF: False,
        }

    @property
    def is_on(self) -> bool | None:
        if self.state is NO_STATE:
            logger.warning('State not set, yet.')
            return None
        return self.state2bool[self.state]

    def validate_state(self, state: str):
        super().validate_state(state)
        if state not in self.state2bool:
            raise InvalidStateValue(component=self, error_msg=f'{state=} not in {self.state2bool.keys()}')

    def get_state(self) -> ComponentState:
        # e.g.: {'topic': 'homeassistant/switch/My-device/My-BinarySensor/state', 'payload': 'ON'}
        return ComponentState(
            topic=f'{self.topic_prefix}/state',
            payload=self.state,
        )

    def get_config(self) -> ComponentConfig:
        return ComponentConfig(
            topic=f'{self.topic_prefix}/config',
            payload={
                'component': self.component,
                'device': self.device.get_mqtt_payload(),
                'device_class': self.device_class,  # e.g.: 'door', 'motion', etc...
                'name': self.name,
                'unique_id': self.uid,
                'payload_off': self.OFF,
                'payload_on': self.ON,
                'state_topic': f'{self.topic_prefix}/state',
                'json_attributes_topic': f'{self.topic_prefix}/attributes',
            },
        )



import typing


if typing.TYPE_CHECKING:
    from ha_services.mqtt4homeassistant.components import BaseComponent


class HaServicesBaseException(Exception):
    """Base class for all exceptions raised by ha_services."""


class InvalidStateValue(HaServicesBaseException):
    def __init__(self, *, component: 'BaseComponent', error_msg: str):
        self.component = component
        self.error_msg = error_msg
        super().__init__(f'{component}: {error_msg}')



import logging
import tempfile
from pathlib import Path

from bx_py_utils.path import assert_is_file
from cli_base.cli_tools.subprocess_utils import ToolsExecutor, verbose_check_call
from cli_base.cli_tools.verbosity import setup_logging
from cli_base.tyro_commands import TyroVerbosityArgType
from rich import print  # noqa

from ha_services.cli_dev import PACKAGE_ROOT, app


logger = logging.getLogger(__name__)


@app.command
def update_ha_data(verbosity: TyroVerbosityArgType):
    """
    Update Home Assistant data files for custom components.

    Creates a venv in /tmp/, installs `homeassistant` package and runs a script to collect data.
    """
    setup_logging(verbosity=verbosity)

    output_path = PACKAGE_ROOT / 'ha_services/ha_data/data_sensors.py'

    collect_script_path = PACKAGE_ROOT / 'ha_services/ha_data/collect_ha_data.py'
    assert_is_file(collect_script_path)

    temp_path = Path(tempfile.gettempdir())
    temp_venv_path = temp_path / 'ha_services_ha_data_venv'
    print(f'Create virtual env in: {temp_venv_path}')

    tools_executor = ToolsExecutor()

    # Install https://pypi.org/project/homeassistant/ package in a virtual env in /tmp/
    tools_executor.verbose_check_call(
        'uv',
        'venv',
        '--allow-existing',
        temp_venv_path,
    )

    collect_script_venv_path = temp_venv_path / 'collect_ha_data.py'
    if collect_script_venv_path.exists():
        collect_script_venv_path.unlink()
    collect_script_venv_path.symlink_to(collect_script_path)

    verbose_check_call(temp_venv_path / 'bin' / 'python', '-V')
    verbose_check_call(temp_venv_path / 'bin' / 'python', '-m', 'ensurepip')
    verbose_check_call(temp_venv_path / 'bin' / 'python', '-m', 'pip', 'install', 'uv')
    verbose_check_call(temp_venv_path / 'bin' / 'uv', 'pip', 'install', 'homeassistant')
    verbose_check_call(
        temp_venv_path / 'bin' / 'python',
        'collect_ha_data.py',
        '--output',
        output_path,
        cwd=temp_venv_path,
    )
    tools_executor.verbose_check_call('ruff', 'format', output_path)



import dataclasses
import logging
import socket
from unittest import TestCase

from ha_services.mqtt4homeassistant.data_classes import MqttSettings
from ha_services.mqtt4homeassistant.utilities.string_utils import slugify
from ha_services.tests.base import ComponentTestMixin


class DataClassesTestCase(ComponentTestMixin, TestCase):
    def test_mqtt_settings(self):
        with self.assertNoLogs(level=logging.ERROR):
            mqtt_settings = MqttSettings()
        self.assertEqual(
            dataclasses.asdict(mqtt_settings),
            {
                'host': 'mqtt.eclipseprojects.io',
                'port': 1883,
                'password': '',
                'user_name': '',
                'main_uid': slugify(socket.gethostname(), sep='_'),
                'publish_config_throttle_seconds': 20,
                'publish_throttle_seconds': 5,
            },
        )



import logging
from unittest import TestCase

from ha_services.exceptions import InvalidStateValue
from ha_services.mqtt4homeassistant.components.select import Select
from ha_services.mqtt4homeassistant.data_classes import ComponentConfig, ComponentState
from ha_services.mqtt4homeassistant.device import MqttDevice


class SelectTestCase(TestCase):
    def test_select(self):
        with self.assertNoLogs(level=logging.WARNING):
            select = Select(
                device=MqttDevice(name='My Device', uid='device1'),
                name='My Select',
                uid='select1',
                options=('Option 1', 'Option 2'),
                default_option='Option 1',
            )
            self.assertEqual(
                select.get_config(),
                ComponentConfig(
                    topic='homeassistant/select/device1/device1-select1/config',
                    payload={
                        'command_topic': 'homeassistant/select/device1/device1-select1/command',
                        'component': 'select',
                        'device': {'identifiers': 'device1', 'name': 'My Device'},
                        'json_attributes_topic': 'homeassistant/select/device1/device1-select1/attributes',
                        'name': 'My Select',
                        'options': ('Option 1', 'Option 2'),
                        'state_topic': 'homeassistant/select/device1/device1-select1/state',
                        'unique_id': 'device1-select1',
                    },
                ),
            )
            self.assertEqual(
                select.get_state(),
                ComponentState(topic='homeassistant/select/device1/device1-select1/state', payload='Option 1'),
            )
            with self.assertRaises(InvalidStateValue) as cm:
                select.set_state('invalid_state')
            self.assertEqual(
                str(cm.exception),
                (
                    "Select(self.uid='device1-select1'):"
                    " state='invalid_state' not in self.options=('Option 1', 'Option 2')"
                ),
            )

            select.set_state('Option 2')
            self.assertEqual(
                select.get_state(),
                ComponentState(topic='homeassistant/select/device1/device1-select1/state', payload='Option 2'),
            )



from bx_py_utils.auto_doc import assert_readme_block
from bx_py_utils.path import assert_is_file
from cli_base.cli_tools.test_utils.rich_test_utils import NoColorEnvRich, invoke
from manageprojects.tests.base import BaseTestCase

from ha_services import constants
from ha_services.cli_dev import PACKAGE_ROOT


def assert_cli_help_in_readme(text_block: str, marker: str):
    README_PATH = PACKAGE_ROOT / 'README.md'
    assert_is_file(README_PATH)

    text_block = text_block.replace(constants.CLI_EPILOG, '')
    text_block = f'```\n{text_block.strip()}\n```'
    assert_readme_block(
        readme_path=README_PATH,
        text_block=text_block,
        start_marker_line=f'[comment]: <> (✂✂✂ auto generated {marker} start ✂✂✂)',
        end_marker_line=f'[comment]: <> (✂✂✂ auto generated {marker} end ✂✂✂)',
    )


class ReadmeTestCase(BaseTestCase):

    def test_main_help(self):
        with NoColorEnvRich():
            stdout = invoke(cli_bin=PACKAGE_ROOT / 'cli.py', args=['--help'], strip_line_prefix='usage: ')
        self.assert_in_content(
            got=stdout,
            parts=(
                'usage: ./cli.py [-h]',
                ' version ',
                'Print version and exit',
                constants.CLI_EPILOG,
            ),
        )
        assert_cli_help_in_readme(text_block=stdout, marker='main help')

    def test_dev_help(self):
        with NoColorEnvRich():
            stdout = invoke(cli_bin=PACKAGE_ROOT / 'dev-cli.py', args=['--help'], strip_line_prefix='usage: ')
        self.assert_in_content(
            got=stdout,
            parts=(
                'usage: ./dev-cli.py [-h]',
                ' lint ',
                ' coverage ',
                ' update-readme-history ',
                ' publish ',
                constants.CLI_EPILOG,
            ),
        )
        assert_cli_help_in_readme(text_block=stdout, marker='dev help')



import datetime
from unittest.mock import patch

from bx_py_utils.test_utils.context_managers import MassContextManager

from ha_services.mqtt4homeassistant.components import get_origin_data
from ha_services.mqtt4homeassistant.mocks.psutil_mock import PsutilMock
from ha_services.mqtt4homeassistant.system_info import wifi_info


IWCONFIG_MOCK_OUTPUT = """
lo        no wireless extensions.

docker0   no wireless extensions.

wlo1      IEEE 802.11  ESSID:"foobar"
          Mode:Managed  Frequency:5.18 GHz  Access Point: 12:34:56:78:AB:CD
          Bit Rate=29.2 Mb/s   Tx-Power=22 dBm
          Retry short limit:7   RTS thr:off   Fragment thr:off
          Power Management:on
          Link Quality=45/70  Signal level=-65 dBm
          Rx invalid nwid:0  Rx invalid crypt:0  Rx invalid frag:0
          Tx excessive retries:0  Invalid misc:511   Missed beacon:0
"""


class HostSystemMock(MassContextManager):
    def __init__(self):

        psutil_mock = PsutilMock()

        origin_data = get_origin_data()
        origin_data['name'] = 'ha-services-tests'
        origin_data['sw_version'] = '1.2.3'

        self.mocks = (
            patch('ha_services.mqtt4homeassistant.components.get_origin_data', return_value=origin_data),
            patch('ha_services.mqtt4homeassistant.device.socket.gethostname', return_value='TheHostName'),
            #
            patch('ha_services.mqtt4homeassistant.system_info.cpu.psutil', psutil_mock),
            patch('ha_services.mqtt4homeassistant.system_info.memory.psutil', psutil_mock),
            patch('ha_services.mqtt4homeassistant.system_info.temperatures.psutil', psutil_mock),
            patch('ha_services.mqtt4homeassistant.utilities.system_utils.psutil', psutil_mock),
            #
            patch.object(wifi_info, 'verbose_check_output', return_value=IWCONFIG_MOCK_OUTPUT),
            #
            # https://github.com/spulec/freezegun/issues/472
            patch("freezegun.api.tzlocal", lambda: datetime.UTC),
        )



import datetime
from unittest import TestCase

from freezegun import freeze_time

from ha_services.mqtt4homeassistant.mocks import HostSystemMock
from ha_services.mqtt4homeassistant.utilities.system_utils import get_system_start_datetime, process_start_datetime


class SystemUtilsTestCase(TestCase):

    def test_get_system_start_datetime(self):
        start_dt = get_system_start_datetime()
        self.assertIsInstance(start_dt, datetime.datetime)
        with (
            HostSystemMock(),
            freeze_time(time_to_freeze='2012-01-14T12:00:00+00:00'),
        ):
            iso_format = get_system_start_datetime().isoformat()
        self.assertEqual(iso_format, '2009-02-13T23:31:30+00:00')

    def test_process_start_datetime(self):
        start_dt = process_start_datetime()
        self.assertIsInstance(start_dt, datetime.datetime)

        with (
            HostSystemMock(),
            freeze_time(time_to_freeze='2012-01-14T12:00:00+00:00'),
        ):
            iso_format = process_start_datetime().isoformat()
        self.assertEqual(iso_format, '2009-02-13T23:31:30+00:00')



import logging
from unittest import TestCase

from frozendict import frozendict

from ha_services.mqtt4homeassistant.components.switch import Switch
from ha_services.mqtt4homeassistant.device import MainMqttDevice, MqttDevice
from ha_services.tests.base import ComponentTestMixin


class DeviceTestCase(ComponentTestMixin, TestCase):
    def test_device(self):
        with self.assertNoLogs(level=logging.WARNING):
            device = MqttDevice(name='My Device', uid='device1')
            self.assertEqual(
                device.get_mqtt_payload(),
                {'name': 'My Device', 'identifiers': 'device1'},
            )
            self.assertEqual(device.topic_prefix, 'homeassistant')
            self.assertEqual(device.uid, 'device1')

            switch1 = Switch(device=device, name='My Switch 1', uid='switch1')
            switch2 = Switch(device=device, name='My Switch 2', uid='switch2')

            with self.assertRaises(AssertionError) as cm:
                Switch(device=device, name='My Switch 1', uid='switch1')
            self.assertEqual(str(cm.exception), 'Duplicate component: device1-switch1')

            self.assertEqual(
                device.components,
                {
                    'device1-switch1': switch1,
                    'device1-switch2': switch2,
                },
            )

    def test_device_with_main_device(self):
        with self.assertNoLogs(level=logging.WARNING):
            main_device = MainMqttDevice(name='Main Device', uid='main_device')
            device = MqttDevice(main_device=main_device, name='Sub Device', uid='sub_device')
            self.assertEqual(
                device.get_mqtt_payload(),
                {
                    'name': 'Sub Device',
                    'identifiers': 'main_device-sub_device',
                    'via_device': 'main_device',
                },
            )
            self.assertEqual(device.topic_prefix, 'homeassistant')
            self.assertEqual(device.uid, 'main_device-sub_device')

    def test_device_extras(self):
        with self.assertNoLogs(level=logging.WARNING):
            device = MqttDevice(
                name='My Device',
                uid='device1',
                manufacturer='Tinkerforge',
                model='HAT Zero Brick',
                sw_version='1.2.3',
            )
            mqtt_payload = device.get_mqtt_payload()
            self.assertEqual(
                mqtt_payload,
                {
                    'name': 'My Device',
                    'identifiers': 'device1',
                    'manufacturer': 'Tinkerforge',
                    'model': 'HAT Zero Brick',
                    'sw_version': '1.2.3',
                },
            )
            self.assertIsInstance(mqtt_payload, frozendict)



"""
    ha_services
    Helpers to send periodic information via MQTT to Home Assistant
"""

# See https://packaging.python.org/en/latest/specifications/version-specifiers/
__version__ = '2.14.0'
__author__ = 'Jens Diemer <github@jensdiemer.de>'



import datetime

from ha_services.mqtt4homeassistant.components.sensor import Sensor
from ha_services.mqtt4homeassistant.utilities.system_utils import get_system_start_datetime, process_start_datetime


class UpTimeSensor(Sensor):
    """
    Sensor for the system up time.
    Adds the datetime when the system was started to Home Assistant.
    """

    def __init__(self, **kwargs):
        kwargs.setdefault('name', 'System Up Time')
        kwargs.setdefault('uid', 'up_time')
        super().__init__(**kwargs)

        # Set only one time. No need to update.
        system_start: datetime.datetime = get_system_start_datetime()
        self.set_state(system_start.isoformat())


class StartTimeSensor(Sensor):
    """
    Sensor for project start time.
    """

    def __init__(self, **kwargs):
        # https://www.home-assistant.io/integrations/sensor/#device-class
        kwargs.setdefault('name', 'Process Start')
        kwargs.setdefault('uid', 'process_start')
        super().__init__(**kwargs)

        # Set only one time. No need to update.
        process_start: datetime.datetime = process_start_datetime()
        self.set_state(process_start.isoformat())



import logging

from cli_base.cli_tools.verbosity import setup_logging
from cli_base.systemd.api import ServiceControl
from cli_base.tyro_commands import TyroVerbosityArgType
from rich import print  # noqa

from ha_services.cli_app import app
from ha_services.cli_app.settings import get_user_settings
from ha_services.example import DemoSettings, SystemdServiceInfo


logger = logging.getLogger(__name__)


def _get_service_control(verbosity: TyroVerbosityArgType) -> ServiceControl:
    setup_logging(verbosity=verbosity)
    user_settings: DemoSettings = get_user_settings(debug=True)
    systemd_settings: SystemdServiceInfo = user_settings.systemd

    return ServiceControl(info=systemd_settings)


@app.command
def systemd_debug(verbosity: TyroVerbosityArgType):
    """
    Print Systemd service template + context + rendered file content.
    """
    service_control = _get_service_control(verbosity)
    service_control.debug_systemd_config()


@app.command
def systemd_setup(verbosity: TyroVerbosityArgType):
    """
    Write Systemd service file, enable it and (re-)start the service. (May need sudo)
    """
    service_control = _get_service_control(verbosity)
    service_control.setup_and_restart_systemd_service()


@app.command
def systemd_remove(verbosity: TyroVerbosityArgType):
    """
    Write Systemd service file, enable it and (re-)start the service. (May need sudo)
    """
    service_control = _get_service_control(verbosity)
    service_control.remove_systemd_service()


@app.command
def systemd_status(verbosity: TyroVerbosityArgType):
    """
    Display status of systemd service. (May need sudo)
    """
    service_control = _get_service_control(verbosity)
    service_control.status()


@app.command
def systemd_stop(verbosity: TyroVerbosityArgType):
    """
    Stops the systemd service. (May need sudo)
    """
    service_control = _get_service_control(verbosity)
    service_control.stop()


@app.command
def systemd_logs(verbosity: TyroVerbosityArgType):
    """
    List and follow logs of systemd service. (May need sudo)
    """
    service_control = _get_service_control(verbosity)
    service_control.logs()



import logging
import time
import typing

from paho.mqtt.client import Client
from psutil._common import snetio

from ha_services.mqtt4homeassistant.components.sensor import Sensor
from ha_services.mqtt4homeassistant.utilities.string_utils import slugify
from ha_services.mqtt4homeassistant.utilities.system_utils import netstat


if typing.TYPE_CHECKING:
    from ha_services.mqtt4homeassistant.device import MqttDevice


logger = logging.getLogger(__name__)


class NetStatSensor:
    def __init__(self, *, device: 'MqttDevice', interface_name: str):
        logger.info('Creating net stat sensor for: %r', interface_name)

        self.device = device
        self.interface_name = interface_name

        self.bytes_sent_old = 0
        self.bytes_sent_sensor = Sensor(
            device=self.device,
            name=f'{interface_name} sent',
            uid=slugify(f'{interface_name}_sent'),
            device_class='data_size',
            state_class='measurement',
            unit_of_measurement='KiB',
            suggested_display_precision=1,
        )
        self.bytes_sent_rate_sensor = Sensor(
            device=self.device,
            name=f'{interface_name} sent rate',
            uid=slugify(f'{interface_name}_sent_rate'),
            device_class='data_rate',
            state_class='measurement',
            unit_of_measurement='KiB/s',
            suggested_display_precision=1,
        )

        self.bytes_received_old = 0
        self.bytes_received_sensor = Sensor(
            device=self.device,
            name=f'{interface_name} received',
            uid=slugify(f'{interface_name}_received'),
            device_class='data_size',
            state_class='measurement',
            unit_of_measurement='KiB',
            suggested_display_precision=1,
        )
        self.bytes_received_rate_sensor = Sensor(
            device=self.device,
            name=f'{interface_name} received rate',
            uid=slugify(f'{interface_name}_received_rate'),
            device_class='data_rate',
            state_class='measurement',
            unit_of_measurement='KiB/s',
            suggested_display_precision=1,
        )
        self.last_update = time.monotonic()

    def set_state(self, data: snetio) -> None:
        update_duration = time.monotonic() - self.last_update
        self.last_update = time.monotonic()

        bytes_sent = data.bytes_sent / 1024
        self.bytes_sent_sensor.set_state(bytes_sent)
        bytes_sent_rate = (bytes_sent - self.bytes_sent_old) / update_duration
        self.bytes_sent_rate_sensor.set_state(bytes_sent_rate)
        self.bytes_sent_old = bytes_sent

        bytes_received = data.bytes_recv / 1024
        self.bytes_received_sensor.set_state(bytes_received)
        bytes_received_rate = (bytes_received - self.bytes_received_old) / update_duration
        self.bytes_received_rate_sensor.set_state(bytes_received_rate)
        self.bytes_received_old = bytes_received

    def publish(self, client):
        self.bytes_sent_sensor.publish(client)
        self.bytes_sent_rate_sensor.publish(client)
        self.bytes_received_sensor.publish(client)
        self.bytes_received_rate_sensor.publish(client)


class NetStatSensors:
    """
    Expose sent/received value from all interesting network interfaces as MQTT sensors.
    """

    def __init__(self, device: 'MqttDevice'):
        self.device = device
        self.sensors = {}
        self.set_state()  # Add sensors

    def set_state(self):
        data = netstat()
        for interface_name, interface_data in data.items():
            try:
                sensor = self.sensors[interface_name]
            except KeyError:
                sensor = NetStatSensor(
                    device=self.device,
                    interface_name=interface_name,
                )
                self.sensors[interface_name] = sensor
            sensor.set_state(interface_data)

    def publish(self, client: Client) -> None:
        self.set_state()
        for sensors in self.sensors.values():
            sensors.publish(client)



"""
    CLI for usage
"""

import logging
import sys
from collections.abc import Sequence

from cli_base.autodiscover import import_all_files
from cli_base.cli_tools.version_info import print_version
from rich import print  # noqa
from tyro.extras import SubcommandApp

import ha_services
from ha_services import constants


logger = logging.getLogger(__name__)

app = SubcommandApp()

# Register all CLI commands, just by import all files in this package:
import_all_files(package=__package__, init_file=__file__)


@app.command
def version():
    """Print version and exit"""
    # Pseudo command, because the version always printed on every CLI call ;)
    sys.exit(0)


def main(args: Sequence[str] | None = None):
    print_version(ha_services)
    app.cli(
        prog='./cli.py',
        description=constants.CLI_EPILOG,
        use_underscores=False,  # use hyphens instead of underscores
        sort_subcommands=True,
        args=args,
    )



import logging

from cli_base.cli_tools.verbosity import setup_logging
from cli_base.tyro_commands import TyroVerbosityArgType
from rich import print  # noqa

from ha_services.cli_app import app
from ha_services.cli_app.settings import get_user_settings
from ha_services.example import DemoSettings, publishing
from ha_services.mqtt4homeassistant.data_classes import MqttSettings
from ha_services.mqtt4homeassistant.mqtt import get_connected_client
from ha_services.mqtt4homeassistant.utilities.debug import print_components


logger = logging.getLogger(__name__)


@app.command
def test_mqtt_connection(verbosity: TyroVerbosityArgType):
    """
    Test connection to MQTT Server
    """
    setup_logging(verbosity=verbosity)
    user_settings: DemoSettings = get_user_settings(debug=True)

    settings: MqttSettings = user_settings.mqtt
    mqttc = get_connected_client(settings=settings, verbosity=verbosity)
    mqttc.loop_start()
    mqttc.loop_stop()
    mqttc.disconnect()
    print('\n[green]Test succeed[/green], bye ;)')


@app.command
def publish_loop(verbosity: TyroVerbosityArgType):
    """
    Publish data via MQTT for Home Assistant (endless loop)
    """
    setup_logging(verbosity=verbosity)
    user_settings: DemoSettings = get_user_settings(debug=True)

    try:
        publishing(user_settings=user_settings, verbosity=verbosity, endless_loop=True)
    except KeyboardInterrupt:
        print('Bye, bye')


@app.command
def info(verbosity: TyroVerbosityArgType):
    """
    Publish data once and display information about all registered components.
    """
    setup_logging(verbosity=verbosity)
    user_settings: DemoSettings = get_user_settings(debug=True)

    try:
        main_device = publishing(user_settings=user_settings, verbosity=verbosity, endless_loop=False)
    except KeyboardInterrupt:
        print('Bye, bye')
    else:
        print_components(main_device)



import logging
import socket

from frozendict import frozendict
from paho.mqtt.client import Client

from ha_services.mqtt4homeassistant.components.sensor import Sensor
from ha_services.mqtt4homeassistant.system_info.cpu import (
    CpuFreqSensor,
    ProcessCpuUsageSensor,
    SystemLoad1MinSensor,
    TotalCpuUsageSensor,
)
from ha_services.mqtt4homeassistant.system_info.memory import SwapUsageSensor
from ha_services.mqtt4homeassistant.system_info.netstat import NetStatSensors
from ha_services.mqtt4homeassistant.system_info.temperatures import TemperaturesSensors
from ha_services.mqtt4homeassistant.system_info.up_time import StartTimeSensor, UpTimeSensor
from ha_services.mqtt4homeassistant.system_info.wifi_info import WifiInfo2Mqtt
from ha_services.mqtt4homeassistant.utilities.assertments import assert_uid


logger = logging.getLogger(__name__)


class BaseMqttDevice:
    device_uids = set()
    components = {}  # Global registry of all components

    def __init__(
        self,
        *,
        name: str,
        uid: str,
        topic_prefix: str = 'homeassistant',
        manufacturer: str | None = None,
        model: str | None = None,
        sw_version: str | None = None,
        throttle_sec: int = 1,  # min. time between state publishing
        config_throttle_sec: int = 20,  # Min. time between config publishing
    ):
        self.name = name

        assert_uid(uid)
        assert uid not in MqttDevice.device_uids, f'Duplicate uid: {uid}'
        self.uid = uid
        self.topic_prefix = topic_prefix

        self.manufacturer = manufacturer
        self.model = model
        self.sw_version = sw_version

        self.throttle_sec = throttle_sec
        self.config_throttle_sec = config_throttle_sec

        self._mqtt_payload_cache = None

    def register_component(self, *, component):
        logger.debug(f'Registering component: {component}')

        from ha_services.mqtt4homeassistant.components import BaseComponent

        assert isinstance(component, BaseComponent)
        uid = component.uid
        assert uid not in self.components, f'Duplicate component: {uid}'
        self.components[uid] = component


class MqttDevice(BaseMqttDevice):
    def __init__(self, *, main_device: BaseMqttDevice | None = None, **kwargs):
        super().__init__(**kwargs)

        if main_device:
            self.via_device = main_device.uid
            self.uid = f'{main_device.uid}-{self.uid}'

    def get_mqtt_payload(self) -> dict:
        if self._mqtt_payload_cache is None:
            mqtt_payload = {
                'name': self.name,
                'identifiers': self.uid,
            }
            for key in ('via_device', 'manufacturer', 'model', 'sw_version'):
                if value := getattr(self, key, None):
                    mqtt_payload[key] = value
            self._mqtt_payload_cache = frozendict(mqtt_payload)

        return self._mqtt_payload_cache


class MainMqttDevice(MqttDevice):
    def __init__(self, **kwargs):
        assert 'main_device' not in kwargs, 'main_device is not allowed for MainMqttDevice'
        super().__init__(**kwargs)

        self.hostname = Sensor(
            device=self,
            name='Hostname',
            uid='hostname',
        )

        self.up_time_sensor = UpTimeSensor(device=self)
        self.process_start_sensor = StartTimeSensor(device=self)
        self.cpu_freq_sensor = CpuFreqSensor(device=self)
        self.swap_usage = SwapUsageSensor(device=self)

        self.system_load_1min = SystemLoad1MinSensor(device=self)
        self.total_cpu_usage = TotalCpuUsageSensor(device=self)
        self.process_cpu_usage = ProcessCpuUsageSensor(device=self)

        self.temperatures_sensors = TemperaturesSensors(device=self)
        self.netstat_sensors = NetStatSensors(device=self)

        self.wifi_info_sensors = WifiInfo2Mqtt(device=self)

    def poll_and_publish(self, client: Client) -> None:
        logger.debug(f'Polling {self.name} ({self.uid})')

        self.hostname.set_state(socket.gethostname())
        self.hostname.publish(client)

        self.up_time_sensor.publish(client)
        self.process_start_sensor.publish(client)
        self.cpu_freq_sensor.publish(client)
        self.swap_usage.publish(client)

        self.system_load_1min.publish(client)
        self.total_cpu_usage.publish(client)
        self.process_cpu_usage.publish(client)

        self.temperatures_sensors.publish(client)
        self.netstat_sensors.publish(client)

        self.wifi_info_sensors.poll_and_publish(client)



import logging
import statistics
import typing

import psutil
from paho.mqtt.client import Client

from ha_services.mqtt4homeassistant.components.sensor import Sensor
from ha_services.mqtt4homeassistant.utilities.string_utils import slugify


if typing.TYPE_CHECKING:
    from ha_services.mqtt4homeassistant.device import MqttDevice


logger = logging.getLogger(__name__)


def median_temperatures(temperatures: dict) -> dict:
    result = {}
    for sensor, temps in temperatures.items():
        currents = [temp.current for temp in temps]
        result[sensor] = statistics.median(currents)
    return result


def get_median_temperatures() -> dict:
    temperatures = psutil.sensors_temperatures()
    return median_temperatures(temperatures)


class TemperaturesSensors:
    def __init__(self, device: 'MqttDevice'):
        self.device = device

        self.sensors = {}

        temperatures = psutil.sensors_temperatures()
        for name in temperatures.keys():
            logger.info('Creating temperature sensor: %r', name)
            sensor = Sensor(
                device=self.device,
                name=f'Temperature {name}',
                uid=slugify(f'temperature_{name}'),
                device_class='temperature',
                state_class='measurement',
                unit_of_measurement='°C',
                suggested_display_precision=0,
            )
            self.sensors[name] = sensor

    def publish(self, client: Client) -> None:
        temperatures = get_median_temperatures()
        for name, median_temperature in temperatures.items():
            sensor = self.sensors[name]
            sensor.set_state(median_temperature)
            sensor.publish(client)



import dataclasses
import logging
import re
import typing

from cli_base.cli_tools.subprocess_utils import verbose_check_output
from paho.mqtt.client import Client

from ha_services.mqtt4homeassistant.components.sensor import Sensor


if typing.TYPE_CHECKING:
    from ha_services.mqtt4homeassistant.device import MqttDevice


logger = logging.getLogger(__name__)


RE_PATTERNS = (
    r'\s+ESSID:"(?P<ESSID>.+)"\s+',
    r'\s+Frequency:(?P<frequency>[\d\.]+?)\s+(?P<frequency_unit>[a-zA-Z]+)\s+',
    r'\s+Bit Rate=(?P<bit_rate>[\d\.]+?)\s+(?P<bit_rate_unit>[a-zA-Z/]+)\s+',
    r'\s+Link Quality=(?P<link_quality>[\d\/]+?)\s+',
    r'\s+Signal level=(?P<signal_level>[-\d\.]+?)\s+(?P<signal_level_unit>[a-zA-Z/]+)\s+',
)


def _get_iwconfig_values() -> dict:
    output = verbose_check_output('iwconfig', verbose=False)

    logger.debug('RAW iwconfig output: %s', output)

    iwconfig_values = {}

    blocks = output.split('\n\n')
    for block in blocks:
        if 'no wireless extensions' in block:
            continue
        if 'ESSID' not in block:
            continue

        device_name = block.partition(' ')[0]

        values = {}
        for re_pattern in RE_PATTERNS:
            match = re.search(re_pattern, block, flags=re.IGNORECASE | re.MULTILINE)
            if match:
                groupdict = match.groupdict()
                values.update(groupdict)

        iwconfig_values[device_name] = values

    return iwconfig_values


@dataclasses.dataclass
class WifiInfoValue:
    slug: str
    name: str
    value: str | float | int
    unit: str | None


@dataclasses.dataclass
class WifiInfo:
    device_name: str
    values: list[WifiInfoValue]


def _convert_iwconfig_values(iwconfig_values: dict) -> list[WifiInfo]:
    wifi_infos = []
    for device_name, values in iwconfig_values.items():
        device_values = []
        for slug, value in sorted(values.items()):
            if slug.endswith('_unit'):
                continue

            if value.isdigit():  # e.g. '123' -> 123
                value = int(value)
            elif re.match(r'\d+\.\d+', value):  # e.g. '1.23' -> 1.23
                value = float(value)
            elif re.match(r'-\d+', value):  # e.g. '-12' -> -12
                value = int(value)

            device_values.append(
                WifiInfoValue(
                    slug=slug,
                    name=slug.replace('_', ' ').capitalize(),  # e.g. 'bit_rate' -> 'Bit rate'
                    value=value,
                    unit=values.get(f'{slug}_unit'),
                )
            )

        wifi_infos.append(
            WifiInfo(
                device_name=device_name,
                values=device_values,
            )
        )

    return wifi_infos


def get_wifi_infos() -> list[WifiInfo]:
    raw_values = _get_iwconfig_values()
    logger.debug('Raw iwconfig values: %r', raw_values)

    wifi_infos = _convert_iwconfig_values(raw_values)
    logger.info('iwconfig values: %r', wifi_infos)
    return wifi_infos


class WifiInfo2Mqtt:
    def __init__(self, *, device: 'MqttDevice'):
        logger.info('Creating WifiInfo sensor')

        self.device = device

        self.wifi_device_name = Sensor(
            device=self.device,
            name='Wifi Device Name',
            uid='wifi_device_name',
        )
        self.sensors = {}

    def poll_and_publish(self, client: Client) -> None:
        wifi_infos = get_wifi_infos()

        device_count = len(wifi_infos)
        if device_count == 0:
            logger.warning('No WiFi devices found')
            return

        if device_count > 1:
            logger.warning('Multiple WiFi devices found, only the first one will be used')

        wifi_info: WifiInfo = wifi_infos[0]

        self.wifi_device_name.set_state(wifi_info.device_name)
        self.wifi_device_name.publish(client)

        for wifi_info_value in wifi_info.values:

            value = wifi_info_value.value  # e.g.: 'foobar', 29.2, ...
            if isinstance(value, (int, float)):
                state_class = 'measurement'
            else:
                state_class = None

            slug = wifi_info_value.slug  # e.g.: 'ESSID', 'bit_rate', ...
            sensor = self.sensors.get(slug)
            if not sensor:
                sensor = self.sensors[slug] = Sensor(
                    device=self.device,
                    name=wifi_info_value.name,  # e.g.: 'ESSID', 'bit rate', ...
                    uid=slug,
                    state_class=state_class,
                    unit_of_measurement=wifi_info_value.unit,  # e.g.: None, 'Mb/s', ...
                )

            sensor.set_state(value)
            sensor.publish(client)



from rich import print  # noqa

from ha_services.mqtt4homeassistant.components.sensor import Sensor
from ha_services.mqtt4homeassistant.device import MainMqttDevice


def print_components(main_device: MainMqttDevice):
    print('\nList all registered components:')

    for index, component in enumerate(main_device.components.values(), start=1):
        print(f'\n{index}. {component.__class__.__name__}:')
        print(f'\tname: [green]{component.name}')
        if isinstance(component, Sensor):
            print(f'\tdevice_class: {component.device_class}')
            print(f'\tstate_class: {component.state_class}')
            state = component.get_state()
            print(f'\tstate: [blue]{state.payload}[/blue] [yellow]{component.unit_of_measurement}')
        print(f'\tthrottle: {component.throttle_sec} seconds')
        print(f'\tconfig throttle: {component.config_throttle_sec} seconds')



import abc
import json
import logging
import time
import typing
from functools import cache

from frozendict import frozendict
from paho.mqtt.client import Client, MQTTMessageInfo

import ha_services
from ha_services.exceptions import InvalidStateValue
from ha_services.mqtt4homeassistant.data_classes import NO_STATE, ComponentConfig, ComponentState, StatePayload
from ha_services.mqtt4homeassistant.utilities.assertments import assert_uid


if typing.TYPE_CHECKING:
    from ha_services.mqtt4homeassistant.device import MqttDevice


logger = logging.getLogger(__name__)


@cache
def get_origin_data() -> dict:
    return {
        'name': 'ha-services',
        'sw_version': ha_services.__version__,
        'support_url': 'https://pypi.org/project/ha_services/',
    }


class BaseComponent(abc.ABC):
    def __init__(
        self,
        *,
        device: 'MqttDevice',
        name: str,
        uid: str,
        component: str,
        initial_state=NO_STATE,  # set_state() must be called to set the value
        qos: int = 0,
        retain: bool = False,
        throttle_sec: int | None = None,  # min. time between state publishing
        config_throttle_sec: int | None = None,  # Min. time between config publishing
    ):
        self.device = device

        self.name = name

        assert_uid(uid)
        self.uid = f'{self.device.uid}-{uid}'

        assert component
        self.component = component
        self.device.register_component(component=self)

        # e.g.: 'homeassistant/sensor/My-device/Chip-Temperature'
        self.topic_prefix = f'{self.device.topic_prefix}/{self.component}/{self.device.uid}/{self.uid}'

        self.state = initial_state
        self.qos = qos
        self.retain = retain

        # Use the given throttle sec values or use them from the device:
        self.throttle_sec = throttle_sec or device.throttle_sec
        self.config_throttle_sec = config_throttle_sec or device.config_throttle_sec
        logger.info(
            f'Initialized {self.topic_prefix} with {self.qos=} {self.retain=}'
            f' {self.throttle_sec=} {self.config_throttle_sec=}'
        )

        self._config_kwargs_cache = None

        # Always publish first config+state:
        self._next_config_publish = 0
        self._next_publish = 0

    def _get_config_kwargs(self) -> dict:
        if self._config_kwargs_cache is None:
            config: ComponentConfig = self.get_config()
            payload = config.payload
            payload['origin'] = get_origin_data()
            payload = json.dumps(config.payload, ensure_ascii=False, sort_keys=True)
            self._config_kwargs_cache = frozendict(
                {
                    'topic': config.topic,
                    'payload': payload,
                    'qos': config.qos,
                    'retain': config.retain,
                }
            )
        return self._config_kwargs_cache

    def publish_config(self, client: Client) -> MQTTMessageInfo | None:
        if self._next_config_publish > time.monotonic():
            logger.debug(f'Publishing {self.uid=} config: throttled')
            return None

        config_kwargs = self._get_config_kwargs()
        logger.info(f'Publishing {self.uid=} config: {config_kwargs}')
        info: MQTTMessageInfo = client.publish(**config_kwargs)

        self._next_config_publish = time.monotonic() + self.config_throttle_sec
        logger.debug(f'Next {self.uid=} config publish: {self._next_config_publish} {self.config_throttle_sec=}')

        return info

    def publish_state(self, client: Client) -> MQTTMessageInfo | None:
        if self.state is NO_STATE:
            logging.warning(f'Sensor {self.uid=} state is not set!')
            return None

        if self._next_publish > time.monotonic():
            logger.debug(f'Publishing {self.uid=}: throttled')
            return None

        state: ComponentState = self.get_state()
        logger.debug(f'Publishing {self.uid=} state: {state}')
        info: MQTTMessageInfo = client.publish(
            topic=state.topic,
            payload=state.payload,
            qos=self.qos,
            retain=self.retain,
        )

        self._next_publish = time.monotonic() + self.throttle_sec

        logger.debug(f'Next {self.uid=} config publish: {self._next_publish} {self.throttle_sec=}')

        return info

    def publish(self, client: Client) -> tuple[MQTTMessageInfo | None, MQTTMessageInfo | None]:
        config_info = self.publish_config(client)
        state_info = self.publish_state(client)
        return config_info, state_info

    @abc.abstractmethod
    def validate_state(self, state: StatePayload):
        """
        raise InvalidStateValue if state is not valid
        """
        if state is NO_STATE:
            raise InvalidStateValue(component=self, error_msg=f'Set {self.uid=} {state=} is not allowed')

    def set_state(self, state: str):
        self.validate_state(state)
        logger.debug('Set state %r for %r', state, self.uid)
        self.state = state

    @abc.abstractmethod
    def get_state(self) -> ComponentState:
        pass

    @abc.abstractmethod
    def get_config(self) -> ComponentConfig:
        pass

    def __str__(self):
        return f'{self.__class__.__name__}({self.uid=})'

    def __repr__(self):
        return f'<{self} {id(self)}>'



from unittest import TestCase

from cli_base.cli_tools.git_history import update_readme_history


class ReadmeHistoryTestCase(TestCase):
    def test_readme_history(self):
        update_readme_history(raise_update_error=True)



from cli_base.cli_tools.dev_tools import run_coverage, run_nox, run_unittest_cli
from cli_base.cli_tools.subprocess_utils import verbose_check_call
from cli_base.cli_tools.test_utils.snapshot import UpdateTestSnapshotFiles
from cli_base.tyro_commands import TyroVerbosityArgType

from ha_services.cli_dev import PACKAGE_ROOT, app


@app.command
def mypy(verbosity: TyroVerbosityArgType):
    """Run Mypy (configured in pyproject.toml)"""
    verbose_check_call('mypy', '.', cwd=PACKAGE_ROOT, verbose=verbosity > 0, exit_on_error=True)


@app.command
def update_test_snapshot_files(verbosity: TyroVerbosityArgType):
    """
    Update all test snapshot files (by remove and recreate all snapshot files)
    """
    with UpdateTestSnapshotFiles(root_path=PACKAGE_ROOT, verbose=verbosity > 0):
        # Just recreate them by running tests:
        run_unittest_cli(
            extra_env=dict(
                RAISE_SNAPSHOT_ERRORS='0',  # Recreate snapshot files without error
            ),
            verbose=verbosity > 1,
            exit_after_run=False,
        )


@app.command  # Dummy command
def test():
    """
    Run unittests
    """
    run_unittest_cli()


@app.command  # Dummy command
def coverage():
    """
    Run tests and show coverage report.
    """
    run_coverage()


@app.command  # Dummy "nox" command
def nox():
    """
    Run nox
    """
    run_nox()



import logging
import socket

import paho.mqtt.client as mqtt
from bx_py_utils.anonymize import anonymize
from cli_base.cli_tools.rich_utils import human_error
from rich import print

from ha_services import __version__
from ha_services.mqtt4homeassistant.data_classes import MqttSettings


logger = logging.getLogger(__name__)


def get_client_id():
    hostname = socket.gethostname()
    client_id = f'ha_services v{__version__} on {hostname}'
    return client_id


class OnConnectCallback:
    def __init__(self, verbosity: int):
        self.verbosity = verbosity

    def __call__(self, client, userdata, flags, reason_code, properties):
        if self.verbosity:
            print(f'MQTT broker connect {reason_code=}', end=' ')

        if reason_code == 0:
            if self.verbosity:
                print('[green]OK')
        else:
            print('\n[red]MQTT Connection not successful!')
            print('[yellow]Please check your credentials\n')
            raise RuntimeError(f'MQTT connection {reason_code=} is not 0')

        if self.verbosity:
            print(f'\t{userdata=}')
            print(f'\t{flags=}')


def get_connected_client(settings: MqttSettings, verbosity: int, timeout=10) -> mqtt.Client:
    client_id = get_client_id()

    if verbosity:
        print(f'\nConnect [cyan]{settings.host}:{settings.port}[/cyan] as "[magenta]{client_id}[/magenta]"', end='...')

    socket.setdefaulttimeout(timeout)  # Sadly: Timeout will not used in getaddrinfo()!
    try:
        info = socket.getaddrinfo(settings.host, settings.port)
    except socket.gaierror as err:
        human_error(
            message=f'{err}\n(Hint: Check you host/port settings)',
            title=f'[red]Error get address info from: [cyan]{settings.host}:{settings.port}[/cyan]',
            exception=err,
            exit_code=1,
        )
    else:
        if not info:
            print('[red]Resolve error: No info!')
        elif verbosity:
            print('Host/port test [green]OK')

    mqttc = mqtt.Client(
        mqtt.CallbackAPIVersion.VERSION2,
        client_id=client_id,
    )
    mqttc.on_connect = OnConnectCallback(verbosity=verbosity)
    mqttc.enable_logger(logger=logger)

    if settings.user_name and settings.password:
        if verbosity:
            print(
                f'login with user: {anonymize(settings.user_name)} password:{anonymize(settings.password)}',
                end='...',
            )
        mqttc.username_pw_set(settings.user_name, settings.password)

    mqttc.connect(settings.host, port=settings.port)

    if verbosity:
        print('[green]OK')
    return mqttc



import psutil

from ha_services.mqtt4homeassistant.components.sensor import Sensor


class SwapUsageSensor(Sensor):
    def __init__(self, **kwargs):
        kwargs.setdefault('name', 'Swap usage')
        kwargs.setdefault('uid', 'swap_usage')
        kwargs.setdefault('unit_of_measurement', '%')
        kwargs.setdefault('suggested_display_precision', 1)
        super().__init__(**kwargs)

    def publish_state(self, *args, **kwargs):
        # Update the state, before publishing:
        swap = psutil.swap_memory()
        self.set_state(swap.percent)
        return super().publish_state(*args, **kwargs)



from ha_services.mqtt4homeassistant.device import BaseMqttDevice


class ComponentTestMixin:
    """
    Clear components to avoid side effects in other tests
    """

    def _clear_components(self):
        BaseMqttDevice.components.clear()

    def setUp(self):
        super().setUp()
        self._clear_components()

    def tearDown(self):
        super().tearDown()
        self._clear_components()



"""
Collect Home Assistant sensor data and write to a Python file.
Don't call this script directly, use:
    .../ha-services/dev-cli.py update-ha-data
"""

import argparse
import datetime
import os
from pathlib import Path
from pprint import pp


def main(out):
    out.write('#\n')
    out.write("# DON'T CHANGE THIS FILE MANUALLY\n")
    out.write('#\n')
    out.write('# This file is generated by:\n')
    out.write('# .../ha-services/dev-cli.py update-ha-data\n')
    out.write('#\n')
    out.write(f'# Last Update: {datetime.datetime.now().isoformat()}\n')
    out.write('#\n')
    import homeassistant

    print(f'{homeassistant.__file__=}')

    from homeassistant.const import __version__

    out.write('# Generated with data from Home Assistant package:\n')
    out.write('#\n')
    out.write(f'# homeassistant v{__version__}\n')
    out.write('#\n')
    out.write('\n')

    from homeassistant.components.sensor import DEVICE_CLASS_STATE_CLASSES, DEVICE_CLASS_UNITS

    ha_sensor_data = []
    for device_class, units in DEVICE_CLASS_UNITS.items():
        state_classes = DEVICE_CLASS_STATE_CLASSES[device_class]
        ha_sensor_data.append(
            {
                'device_class': str(device_class),
                'state_classes': set(str(sc) for sc in state_classes),
                'units': set(str(unit) for unit in units if unit),
            }
        )

    out.write(f'HA_SENSOR_DATA = {ha_sensor_data!r}\n')
    pp(ha_sensor_data, width=os.get_terminal_size().columns)
    print(f'\n\nWrote data to: {out.name}')


if __name__ == '__main__':
    print('_' * 100)
    print(__file__)

    parser = argparse.ArgumentParser(description='Collect Home Assistant sensor data.')
    parser.add_argument('--output', type=str, help='Output file path', default='ha_sensor_data.py')
    args = parser.parse_args()

    output_path = Path(args.output)

    with output_path.open('w') as out:
        main(out)



import logging

from cli_base.cli_tools.dev_tools import run_unittest_cli
from cli_base.cli_tools.subprocess_utils import ToolsExecutor
from cli_base.cli_tools.verbosity import setup_logging
from cli_base.run_pip_audit import run_pip_audit
from cli_base.tyro_commands import TyroVerbosityArgType
from manageprojects.utilities.publish import publish_package

import ha_services
from ha_services.cli_dev import PACKAGE_ROOT, app


logger = logging.getLogger(__name__)


@app.command
def install():
    """
    Install requirements and 'ha_services' via pip as editable.
    """
    tools_executor = ToolsExecutor(cwd=PACKAGE_ROOT)
    tools_executor.verbose_check_call('uv', 'sync')
    tools_executor.verbose_check_call('pip', 'install', '--no-deps', '-e', '.')


@app.command
def pip_audit(verbosity: TyroVerbosityArgType):
    """
    Run pip-audit check against current requirements files
    """
    setup_logging(verbosity=verbosity)
    run_pip_audit(base_path=PACKAGE_ROOT, verbosity=verbosity)


@app.command
def update(verbosity: TyroVerbosityArgType):
    """
    Update dependencies (uv.lock) and git pre-commit hooks
    """
    setup_logging(verbosity=verbosity)

    tools_executor = ToolsExecutor(cwd=PACKAGE_ROOT)

    tools_executor.verbose_check_call('pip', 'install', '-U', 'pip')
    tools_executor.verbose_check_call('pip', 'install', '-U', 'uv')
    tools_executor.verbose_check_call('uv', 'lock', '--upgrade')

    run_pip_audit(base_path=PACKAGE_ROOT, verbosity=verbosity)

    # Install new dependencies in current .venv:
    tools_executor.verbose_check_call('uv', 'sync')

    # Update git pre-commit hooks:
    tools_executor.verbose_check_call('pre-commit', 'autoupdate')


@app.command
def publish():
    """
    Build and upload this project to PyPi
    """
    run_unittest_cli(verbose=False, exit_after_run=False)  # Don't publish a broken state

    publish_package(module=ha_services, package_path=PACKAGE_ROOT)



#
# DON'T CHANGE THIS FILE MANUALLY
#
# This file is generated by:
# .../ha-services/dev-cli.py update-ha-data
#
# Last Update: 2025-09-25T18:22:29.675522
#
# Generated with data from Home Assistant package:
#
# homeassistant v2025.9.4
#

HA_SENSOR_DATA = [
    {'device_class': 'absolute_humidity', 'state_classes': {'measurement'}, 'units': {'mg/m³', 'g/m³'}},
    {'device_class': 'apparent_power', 'state_classes': {'measurement'}, 'units': {'mVA', 'VA'}},
    {'device_class': 'aqi', 'state_classes': {'measurement'}, 'units': set()},
    {
        'device_class': 'area',
        'state_classes': {'measurement_angle', 'total', 'measurement', 'total_increasing'},
        'units': {'in²', 'mm²', 'yd²', 'ha', 'm²', 'km²', 'mi²', 'cm²', 'ac', 'ft²'},
    },
    {
        'device_class': 'atmospheric_pressure',
        'state_classes': {'measurement'},
        'units': {'mbar', 'psi', 'cbar', 'hPa', 'mmHg', 'inHg', 'kPa', 'bar', 'Pa'},
    },
    {'device_class': 'battery', 'state_classes': {'measurement'}, 'units': {'%'}},
    {'device_class': 'blood_glucose_concentration', 'state_classes': {'measurement'}, 'units': {'mg/dL', 'mmol/L'}},
    {'device_class': 'carbon_monoxide', 'state_classes': {'measurement'}, 'units': {'ppm'}},
    {'device_class': 'carbon_dioxide', 'state_classes': {'measurement'}, 'units': {'ppm'}},
    {'device_class': 'conductivity', 'state_classes': {'measurement'}, 'units': {'S/cm', 'μS/cm', 'mS/cm'}},
    {'device_class': 'current', 'state_classes': {'measurement'}, 'units': {'A', 'mA'}},
    {
        'device_class': 'data_rate',
        'state_classes': {'measurement'},
        'units': {'KiB/s', 'B/s', 'kB/s', 'MB/s', 'kbit/s', 'GiB/s', 'GB/s', 'MiB/s', 'Mbit/s', 'bit/s', 'Gbit/s'},
    },
    {
        'device_class': 'data_size',
        'state_classes': {'measurement_angle', 'total', 'measurement', 'total_increasing'},
        'units': {
            'TB',
            'EB',
            'PB',
            'bit',
            'YiB',
            'kbit',
            'ZiB',
            'KiB',
            'kB',
            'MiB',
            'Mbit',
            'Gbit',
            'ZB',
            'TiB',
            'PiB',
            'EiB',
            'YB',
            'GB',
            'GiB',
            'MB',
            'B',
        },
    },
    {
        'device_class': 'distance',
        'state_classes': {'measurement_angle', 'total', 'measurement', 'total_increasing'},
        'units': {'cm', 'm', 'ft', 'in', 'nmi', 'km', 'mm', 'yd', 'mi'},
    },
    {
        'device_class': 'duration',
        'state_classes': {'measurement_angle', 'total', 'measurement', 'total_increasing'},
        'units': {'min', 'h', 's', 'μs', 'd', 'ms'},
    },
    {
        'device_class': 'energy',
        'state_classes': {'total', 'total_increasing'},
        'units': {'Gcal', 'mWh', 'Wh', 'kWh', 'J', 'Mcal', 'kJ', 'MWh', 'kcal', 'TWh', 'GJ', 'MJ', 'GWh', 'cal'},
    },
    {
        'device_class': 'energy_distance',
        'state_classes': {'measurement'},
        'units': {'km/kWh', 'mi/kWh', 'kWh/100km', 'Wh/km'},
    },
    {
        'device_class': 'energy_storage',
        'state_classes': {'measurement'},
        'units': {'Gcal', 'mWh', 'Wh', 'kWh', 'J', 'Mcal', 'kJ', 'MWh', 'kcal', 'TWh', 'GJ', 'MJ', 'GWh', 'cal'},
    },
    {'device_class': 'frequency', 'state_classes': {'measurement'}, 'units': {'MHz', 'kHz', 'Hz', 'GHz'}},
    {'device_class': 'gas', 'state_classes': {'total', 'total_increasing'}, 'units': {'CCF', 'L', 'm³', 'ft³'}},
    {'device_class': 'humidity', 'state_classes': {'measurement'}, 'units': {'%'}},
    {'device_class': 'illuminance', 'state_classes': {'measurement'}, 'units': {'lx'}},
    {'device_class': 'irradiance', 'state_classes': {'measurement'}, 'units': {'W/m²', 'BTU/(h⋅ft²)'}},
    {'device_class': 'moisture', 'state_classes': {'measurement'}, 'units': {'%'}},
    {'device_class': 'nitrogen_dioxide', 'state_classes': {'measurement'}, 'units': {'μg/m³'}},
    {'device_class': 'nitrogen_monoxide', 'state_classes': {'measurement'}, 'units': {'μg/m³'}},
    {'device_class': 'nitrous_oxide', 'state_classes': {'measurement'}, 'units': {'μg/m³'}},
    {'device_class': 'ozone', 'state_classes': {'measurement'}, 'units': {'μg/m³'}},
    {'device_class': 'ph', 'state_classes': {'measurement'}, 'units': set()},
    {'device_class': 'pm1', 'state_classes': {'measurement'}, 'units': {'μg/m³'}},
    {'device_class': 'pm10', 'state_classes': {'measurement'}, 'units': {'μg/m³'}},
    {'device_class': 'pm25', 'state_classes': {'measurement'}, 'units': {'μg/m³'}},
    {'device_class': 'power_factor', 'state_classes': {'measurement'}, 'units': {'%'}},
    {'device_class': 'power', 'state_classes': {'measurement'}, 'units': {'MW', 'TW', 'GW', 'mW', 'W', 'kW'}},
    {
        'device_class': 'precipitation',
        'state_classes': {'measurement_angle', 'total', 'measurement', 'total_increasing'},
        'units': {'in', 'cm', 'mm'},
    },
    {
        'device_class': 'precipitation_intensity',
        'state_classes': {'measurement'},
        'units': {'in/h', 'in/d', 'mm/d', 'mm/h'},
    },
    {
        'device_class': 'pressure',
        'state_classes': {'measurement'},
        'units': {'mbar', 'psi', 'cbar', 'hPa', 'mmHg', 'inHg', 'kPa', 'bar', 'Pa'},
    },
    {'device_class': 'reactive_energy', 'state_classes': {'total', 'total_increasing'}, 'units': {'varh', 'kvarh'}},
    {'device_class': 'reactive_power', 'state_classes': {'measurement'}, 'units': {'mvar', 'kvar', 'var'}},
    {'device_class': 'signal_strength', 'state_classes': {'measurement'}, 'units': {'dB', 'dBm'}},
    {'device_class': 'sound_pressure', 'state_classes': {'measurement'}, 'units': {'dBA', 'dB'}},
    {
        'device_class': 'speed',
        'state_classes': {'measurement'},
        'units': {'mph', 'mm/s', 'Beaufort', 'mm/h', 'ft/s', 'm/s', 'mm/d', 'km/h', 'in/s', 'in/h', 'in/d', 'kn'},
    },
    {'device_class': 'sulphur_dioxide', 'state_classes': {'measurement'}, 'units': {'μg/m³'}},
    {'device_class': 'temperature', 'state_classes': {'measurement'}, 'units': {'K', '°F', '°C'}},
    {'device_class': 'volatile_organic_compounds', 'state_classes': {'measurement'}, 'units': {'mg/m³', 'μg/m³'}},
    {'device_class': 'volatile_organic_compounds_parts', 'state_classes': {'measurement'}, 'units': {'ppb', 'ppm'}},
    {'device_class': 'voltage', 'state_classes': {'measurement'}, 'units': {'MV', 'kV', 'μV', 'mV', 'V'}},
    {
        'device_class': 'volume',
        'state_classes': {'total', 'total_increasing'},
        'units': {'mL', 'fl. oz.', 'ft³', 'gal', 'm³', 'CCF', 'L'},
    },
    {
        'device_class': 'volume_flow_rate',
        'state_classes': {'measurement'},
        'units': {'L/h', 'm³/h', 'gal/min', 'm³/min', 'ft³/min', 'L/min', 'm³/s', 'mL/s', 'L/s'},
    },
    {
        'device_class': 'volume_storage',
        'state_classes': {'measurement'},
        'units': {'mL', 'fl. oz.', 'ft³', 'gal', 'm³', 'CCF', 'L'},
    },
    {
        'device_class': 'water',
        'state_classes': {'total', 'total_increasing'},
        'units': {'ft³', 'gal', 'm³', 'CCF', 'L'},
    },
    {
        'device_class': 'weight',
        'state_classes': {'total', 'measurement', 'total_increasing'},
        'units': {'lb', 'kg', 'oz', 'mg', 'st', 'g', 'μg'},
    },
    {'device_class': 'wind_direction', 'state_classes': {'measurement_angle'}, 'units': {'°'}},
    {
        'device_class': 'wind_speed',
        'state_classes': {'measurement'},
        'units': {'mph', 'mm/s', 'Beaufort', 'ft/s', 'm/s', 'km/h', 'in/s', 'kn'},
    },
]



import os

import psutil

from ha_services.mqtt4homeassistant.components.sensor import Sensor


class CpuFreqSensor(Sensor):

    def __init__(self, **kwargs):
        kwargs.setdefault('name', 'CPU frequency')
        kwargs.setdefault('uid', 'cpu_freq')
        kwargs.setdefault('device_class', 'frequency')
        kwargs.setdefault('state_class', 'measurement')
        kwargs.setdefault('unit_of_measurement', 'MHz')
        kwargs.setdefault('suggested_display_precision', 0)
        kwargs.setdefault('min_value', 1)  # Less than 1MHz ?
        kwargs.setdefault('max_value', 20_000)  # More than 20.000MHz ?
        super().__init__(**kwargs)

    def publish_state(self, *args, **kwargs):
        # Update the state, before publishing:
        info = psutil.cpu_freq()
        self.set_state(int(info.current))
        return super().publish_state(*args, **kwargs)


class SystemLoad1MinSensor(Sensor):

    def __init__(self, **kwargs):
        kwargs.setdefault('name', 'System load 1min.')
        kwargs.setdefault('uid', 'system_load_1min')
        kwargs.setdefault('state_class', 'measurement')
        kwargs.setdefault('suggested_display_precision', 2)
        kwargs.setdefault('min_value', 0)
        super().__init__(**kwargs)

    def publish_state(self, *args, **kwargs):
        # Update the state, before publishing:
        self.set_state(psutil.getloadavg()[0])
        return super().publish_state(*args, **kwargs)


class TotalCpuUsageSensor(Sensor):

    def __init__(self, **kwargs):
        kwargs.setdefault('name', 'Total CPU usage')
        kwargs.setdefault('uid', 'total_cpu_usage')
        kwargs.setdefault('unit_of_measurement', '%')
        kwargs.setdefault('suggested_display_precision', 1)
        kwargs.setdefault('min_value', 0)
        kwargs.setdefault('max_value', 100)
        super().__init__(**kwargs)

    def publish_state(self, *args, **kwargs):
        # Update the state, before publishing:
        self.set_state(psutil.cpu_percent(interval=None))
        return super().publish_state(*args, **kwargs)


class ProcessCpuUsageSensor(Sensor):

    def __init__(self, **kwargs):
        kwargs.setdefault('name', 'Process CPU usage')
        kwargs.setdefault('uid', 'process_cpu_usage')
        kwargs.setdefault('unit_of_measurement', '%')
        kwargs.setdefault('suggested_display_precision', 1)
        kwargs.setdefault('min_value', 0)
        kwargs.setdefault('max_value', 100)
        super().__init__(**kwargs)

        self.process = psutil.Process(os.getpid())

    def publish_state(self, *args, **kwargs):
        # Update the state, before publishing:
        self.set_state(self.process.cpu_percent(interval=None))
        return super().publish_state(*args, **kwargs)



from psutil._common import shwtemp, snetio


class UsageMock:
    ru_utime = 1
    ru_stime = 1


class CpuFreqMock:
    current = 1234


class SwapMemoryMock:
    percent = 123


class PsutilMock:

    def boot_time(self):
        return 1234567890

    def getloadavg(self):
        return (1, 2, 3)

    def cpu_freq(self):
        return CpuFreqMock

    def cpu_percent(self, interval=None):
        return 12

    def swap_memory(self):
        return SwapMemoryMock

    def sensors_temperatures(self):
        return {
            'coretemp': [
                shwtemp(label='Package id 0', current=53.0, high=100.0, critical=100.0),
                shwtemp(label='Core 0', current=50.0, high=100.0, critical=100.0),
                shwtemp(label='Core 4', current=51.0, high=100.0, critical=100.0),
            ],
            'nvme': [
                shwtemp(label='Composite', current=32.85, high=81.85, critical=84.85),
            ],
        }

    def net_io_counters(self, pernic):
        return {
            'eth0': snetio(
                bytes_sent=123,
                bytes_recv=456,
                packets_sent=0,
                packets_recv=0,
                errin=0,
                errout=0,
                dropin=0,
                dropout=0,
            ),
        }

    def Process(self, pid):
        class ProcessMock:
            def create_time(self):
                return 1234567890

            def cpu_percent(self, interval=None):
                return 12

        return ProcessMock()



"""
    CLI for development
"""

import importlib
import logging
import sys
from collections.abc import Sequence

from bx_py_utils.path import assert_is_file
from cli_base.autodiscover import import_all_files
from cli_base.cli_tools.dev_tools import run_coverage, run_nox, run_unittest_cli
from cli_base.cli_tools.version_info import print_version
from typeguard import install_import_hook
from tyro.extras import SubcommandApp

import ha_services
from ha_services import constants


# Check type annotations via typeguard in all tests.
# Sadly we must activate this here and can't do this in ./tests/__init__.py
install_import_hook(packages=('ha_services',))

# reload the module, after the typeguard import hook is activated:
importlib.reload(ha_services)


logger = logging.getLogger(__name__)


PACKAGE_ROOT = constants.BASE_PATH.parent
assert_is_file(PACKAGE_ROOT / 'pyproject.toml')  # Exists only in cloned git repo


app = SubcommandApp()


# Register all CLI commands, just by import all files in this package:
import_all_files(package=__package__, init_file=__file__)


@app.command
def version():
    """Print version and exit"""
    # Pseudo command, because the version always printed on every CLI call ;)
    sys.exit(0)


def main(args: Sequence[str] | None = None):
    print_version(ha_services)

    if len(sys.argv) >= 2:
        # Check if we can just pass a command call to origin CLI:
        command = sys.argv[1]
        command_map = {
            'test': run_unittest_cli,
            'nox': run_nox,
            'coverage': run_coverage,
        }
        if real_func := command_map.get(command):
            real_func(argv=sys.argv, exit_after_run=True)

    app.cli(
        prog='./dev-cli.py',
        description=constants.CLI_EPILOG,
        use_underscores=False,  # use hyphens instead of underscores
        sort_subcommands=True,
        args=args,
    )



import logging
from unittest import TestCase

from bx_py_utils.test_utils.snapshot import assert_snapshot

from ha_services.mqtt4homeassistant.data_classes import ComponentState
from ha_services.mqtt4homeassistant.device import MqttDevice
from ha_services.mqtt4homeassistant.mocks import HostSystemMock
from ha_services.mqtt4homeassistant.mocks.mqtt_client_mock import MqttClientMock
from ha_services.mqtt4homeassistant.system_info.netstat import NetStatSensor, NetStatSensors


class NetStatSensorsTestCase(TestCase):
    maxDiff = None

    def test_happy_path(self):
        def get_sensors_data(netstat_sensors):
            data = {}
            for interface_name, sensor in netstat_sensors.sensors.items():
                state = sensor.bytes_sent_sensor.get_state()
                self.assertIsInstance(state, ComponentState)
                data[interface_name] = state
            return data

        with HostSystemMock(), self.assertNoLogs(level=logging.WARNING):
            netstat_sensors = NetStatSensors(
                device=MqttDevice(name='foo', uid='bar'),
            )
            self.assertEqual(
                get_sensors_data(netstat_sensors),
                {
                    'eth0': ComponentState(
                        topic='homeassistant/sensor/bar/bar-eth0sent/state',
                        payload=0.1201171875,  # Mocked value
                    )
                },
            )

            sensor = netstat_sensors.sensors['eth0']
            self.assertIsInstance(sensor, NetStatSensor)
            self.assertEqual(
                sensor.bytes_sent_sensor.get_state().topic,
                'homeassistant/sensor/bar/bar-eth0sent/state',
            )

            mqtt_client_mock = MqttClientMock()
            netstat_sensors.publish(mqtt_client_mock)

            self.assertEqual(
                mqtt_client_mock.get_state_messages(),
                [
                    {
                        'topic': 'homeassistant/sensor/bar/bar-eth0sent/state',
                        'payload': 0.1201171875,
                        'qos': 0,
                        'retain': False,
                    },
                    {
                        'topic': 'homeassistant/sensor/bar/bar-eth0sentrate/state',
                        'payload': 0.0,
                        'qos': 0,
                        'retain': False,
                    },
                    {
                        'topic': 'homeassistant/sensor/bar/bar-eth0received/state',
                        'payload': 0.4453125,
                        'qos': 0,
                        'retain': False,
                    },
                    {
                        'topic': 'homeassistant/sensor/bar/bar-eth0receivedrate/state',
                        'payload': 0.0,
                        'qos': 0,
                        'retain': False,
                    },
                ],
            )

            config_payload = mqtt_client_mock.get_config_payload()
            # Check sample:
            self.assertEqual(
                config_payload[0],
                {
                    'component': 'sensor',
                    'device': {'identifiers': 'bar', 'name': 'foo'},
                    'device_class': 'data_size',
                    'json_attributes_topic': 'homeassistant/sensor/bar/bar-eth0sent/attributes',
                    'name': 'eth0 sent',
                    'origin': {
                        'name': 'ha-services-tests',
                        'support_url': 'https://pypi.org/project/ha_services/',
                        'sw_version': '1.2.3',
                    },
                    'state_class': 'measurement',
                    'state_topic': 'homeassistant/sensor/bar/bar-eth0sent/state',
                    'suggested_display_precision': 1,
                    'unique_id': 'bar-eth0sent',
                    'unit_of_measurement': 'KiB',
                },
            )
            assert_snapshot(got=config_payload)

            state_messages = mqtt_client_mock.get_state_messages()
            # Check sample:
            self.assertEqual(
                state_messages[0],
                {
                    'topic': 'homeassistant/sensor/bar/bar-eth0sent/state',
                    'payload': 0.1201171875,
                    'qos': 0,
                    'retain': False,
                },
            )
            assert_snapshot(got=state_messages)



import datetime
import logging
import os

import psutil


logger = logging.getLogger(__name__)


def get_system_start_datetime() -> datetime.datetime:
    tznow = datetime.datetime.now().astimezone()
    start_dt = datetime.datetime.fromtimestamp(psutil.boot_time(), tz=tznow.tzinfo)
    return start_dt


def process_start_datetime() -> datetime.datetime:
    p = psutil.Process(os.getpid())
    create_time: float = p.create_time()
    tznow = datetime.datetime.now().astimezone()
    start_dt = datetime.datetime.fromtimestamp(create_time, tz=tznow.tzinfo)
    return start_dt


def netstat() -> dict:
    netstat = psutil.net_io_counters(pernic=True)
    filtered = {}
    for interface, statistics in netstat.items():
        if interface == 'lo' or interface.startswith(('veth', 'docker', 'br-')):
            logger.debug(f'Skipping network interface: {interface}')
            continue
        logger.debug(f'Including network interface: {interface}')
        filtered[interface] = statistics
    return filtered


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)

    print(f'{get_system_start_datetime().isoformat()=}')
    print(f'{process_start_datetime().isoformat()=}')
    print(f'{netstat()=}')



import os
import unittest.util
from pathlib import Path

from bx_py_utils.test_utils.deny_requests import deny_any_real_request
from cli_base.cli_tools.verbosity import MAX_LOG_LEVEL, setup_logging
from rich import print  # noqa
from typeguard import install_import_hook


# Check type annotations via typeguard in all tests:
install_import_hook(packages=('ha_services',))


def pre_configure_tests() -> None:
    print(f'Configure unittests via "load_tests Protocol" from {Path(__file__).relative_to(Path.cwd())}')

    # Hacky way to display more "assert"-Context in failing tests:
    _MIN_MAX_DIFF = unittest.util._MAX_LENGTH - unittest.util._MIN_DIFF_LEN
    unittest.util._MAX_LENGTH = int(os.environ.get('UNITTEST_MAX_LENGTH', 2000))
    unittest.util._MIN_DIFF_LEN = unittest.util._MAX_LENGTH - _MIN_MAX_DIFF

    # Deny any request via docket/urllib3 because tests they should mock all requests:
    deny_any_real_request()

    # Display DEBUG logs in tests:
    setup_logging(verbosity=MAX_LOG_LEVEL)


def load_tests(loader, tests, pattern):
    """
    Use unittest "load_tests Protocol" as a hook to setup test environment before running tests.
    https://docs.python.org/3/library/unittest.html#load-tests-protocol
    """
    pre_configure_tests()
    return loader.discover(start_dir=Path(__file__).parent, pattern=pattern)



import logging
import sys
from pathlib import Path

from cli_base.cli_tools import git_history
from cli_base.cli_tools.verbosity import setup_logging
from cli_base.tyro_commands import TyroVerbosityArgType
from rich import print  # noqa

from ha_services.cli_app import app


logger = logging.getLogger(__name__)


@app.command
def update_readme_history(verbosity: TyroVerbosityArgType):
    """
    Update project history base on git commits/tags in README.md

    Will be exited with 1 if the README.md was updated otherwise with 0.

    Also, callable via e.g.:
        python -m cli_base update-readme-history -v
    """
    setup_logging(verbosity=verbosity)

    logger.debug('%s called. CWD: %s', __name__, Path.cwd())
    updated = git_history.update_readme_history(verbosity=verbosity)
    exit_code = 1 if updated else 0
    if verbosity:
        print(f'{exit_code=}')
    sys.exit(exit_code)



import logging

from cli_base.cli_tools.shell_completion import setup_tyro_shell_completion
from cli_base.cli_tools.verbosity import setup_logging
from cli_base.tyro_commands import TyroVerbosityArgType
from rich import print  # noqa

from ha_services.cli_dev import app


logger = logging.getLogger(__name__)


@app.command
def shell_completion(verbosity: TyroVerbosityArgType = 1, remove: bool = False) -> None:
    """
    Setup shell completion for this CLI (Currently only for bash shell)
    """
    setup_logging(verbosity=verbosity)
    setup_tyro_shell_completion(
        prog_name='ha_services_dev_cli',
        remove=remove,
    )



from ha_services.ha_data.data_sensors import HA_SENSOR_DATA


HA_SENSOR_MAP = {sensor['device_class']: sensor for sensor in HA_SENSOR_DATA}


class ValidationError(ValueError):
    pass


def validate_sensor(*, device_class: str | None, state_class: str | None, unit_of_measurement: str | None) -> None:
    """
    Validate if the given device_class, state_class and unit_of_measurement are valid according to HA_SENSOR_DATA.
    """

    if device_class is None:
        return

    try:
        ha_data = HA_SENSOR_MAP[device_class]
    except KeyError as err:
        raise ValidationError(
            f'Invalid: {device_class=} ! Valid options: {",".join(sorted(HA_SENSOR_MAP.keys()))}'
        ) from err

    if state_class is not None and state_class not in ha_data['state_classes']:
        raise ValidationError(f'Invalid: {state_class=} for {device_class=}. Valid options: {ha_data["state_classes"]}')

    if unit_of_measurement is not None and unit_of_measurement not in ha_data['units']:
        raise ValidationError(f'Invalid: {unit_of_measurement=} for {device_class=}. Valid options: {ha_data["units"]}')



import unittest

from ha_services.ha_data.validators import ValidationError, validate_sensor


class ValidateSensorTestCase(unittest.TestCase):
    def test_happy_path(self):
        validate_sensor(device_class=None, state_class=None, unit_of_measurement=None)
        validate_sensor(device_class='battery', state_class='measurement', unit_of_measurement='%')

        with self.assertRaises(ValidationError):
            validate_sensor(device_class='invalid_class', state_class=None, unit_of_measurement=None)

        with self.assertRaises(ValidationError):
            validate_sensor(device_class='battery', state_class='invalid_state', unit_of_measurement=None)

        with self.assertRaises(ValidationError):
            validate_sensor(device_class='battery', state_class='measurement', unit_of_measurement='invalid_unit')



import logging
from unittest import TestCase

from ha_services.exceptions import InvalidStateValue
from ha_services.ha_data.validators import ValidationError
from ha_services.mqtt4homeassistant.components.sensor import Sensor
from ha_services.mqtt4homeassistant.data_classes import NO_STATE, ComponentConfig, ComponentState
from ha_services.mqtt4homeassistant.device import MqttDevice
from ha_services.tests.base import ComponentTestMixin


class SensorTestCase(ComponentTestMixin, TestCase):
    def test_mimimal_sensor(self):
        with self.assertNoLogs(level=logging.WARNING):
            sensor = Sensor(
                device=MqttDevice(name='My device', uid='device_id'),
                name='My component',
                uid='component_id',
            )
            self.assertEqual(
                sensor.get_config(),
                ComponentConfig(
                    topic='homeassistant/sensor/device_id/device_id-component_id/config',
                    payload={
                        'component': 'sensor',
                        'device': {'identifiers': 'device_id', 'name': 'My device'},
                        'device_class': None,
                        'json_attributes_topic': 'homeassistant/sensor/device_id/device_id-component_id/attributes',
                        'name': 'My component',
                        'state_class': None,
                        'state_topic': 'homeassistant/sensor/device_id/device_id-component_id/state',
                        'unique_id': 'device_id-component_id',
                        'unit_of_measurement': None,
                    },
                ),
            )
            self.assertEqual(
                sensor.get_state(),
                ComponentState(
                    topic='homeassistant/sensor/device_id/device_id-component_id/state',
                    payload=NO_STATE,  # no initial state -> NO_STATE
                ),
            )
            sensor.set_state(123)
            self.assertEqual(
                sensor.get_state(),
                ComponentState(
                    topic='homeassistant/sensor/device_id/device_id-component_id/state',
                    payload=123,  # now we have a state
                ),
            )

    def test_min_max_validation(self):
        with self.assertNoLogs(level=logging.WARNING):
            sensor = Sensor(
                device=MqttDevice(name='My device', uid='device_id'),
                name='My component',
                uid='component_id',
                min_value=-10,
                max_value=10,
            )
            self.assertEqual(
                sensor.get_config(),
                ComponentConfig(
                    topic='homeassistant/sensor/device_id/device_id-component_id/config',
                    payload={
                        'component': 'sensor',
                        'device': {'identifiers': 'device_id', 'name': 'My device'},
                        'device_class': None,
                        'json_attributes_topic': 'homeassistant/sensor/device_id/device_id-component_id/attributes',
                        'name': 'My component',
                        'state_class': None,
                        'state_topic': 'homeassistant/sensor/device_id/device_id-component_id/state',
                        'unique_id': 'device_id-component_id',
                        'unit_of_measurement': None,
                    },
                ),
            )
            self.assertEqual(
                sensor.get_state(),
                ComponentState(
                    topic='homeassistant/sensor/device_id/device_id-component_id/state',
                    payload=NO_STATE,  # no initial state -> NO_STATE
                ),
            )

            with self.assertRaises(InvalidStateValue) as cm:
                sensor.set_state(-123.456)
            self.assertEqual(
                str(cm.exception),
                "Sensor(self.uid='device_id-component_id'): state=-123.456 is smaller than self.min_value=-10",
            )

            with self.assertRaises(InvalidStateValue) as cm:
                sensor.set_state('Bam!')
            self.assertEqual(
                str(cm.exception), "Sensor(self.uid='device_id-component_id'): state='Bam!' is not a number"
            )

            with self.assertRaises(InvalidStateValue) as cm:
                sensor.set_state(123)
            self.assertEqual(
                str(cm.exception),
                "Sensor(self.uid='device_id-component_id'): state=123 is bigger than self.max_value=10",
            )

            sensor.set_state(5)
            self.assertEqual(
                sensor.get_state(),
                ComponentState(
                    topic='homeassistant/sensor/device_id/device_id-component_id/state',
                    payload=5,  # now we have a state
                ),
            )

    def test_full_sensor(self):
        with self.assertNoLogs(level=logging.WARNING):
            sensor = Sensor(
                device=MqttDevice(name='My device', uid='device_id'),
                name='My component',
                uid='component_id',
                component='sensor',
                device_class='temperature',
                state_class='measurement',
                unit_of_measurement='°C',
                suggested_display_precision=2,
            )
            self.assertEqual(
                sensor.get_config(),
                ComponentConfig(
                    topic='homeassistant/sensor/device_id/device_id-component_id/config',
                    payload={
                        'component': 'sensor',
                        'device': {'identifiers': 'device_id', 'name': 'My device'},
                        'device_class': 'temperature',
                        'json_attributes_topic': 'homeassistant/sensor/device_id/device_id-component_id/attributes',
                        'name': 'My component',
                        'state_class': 'measurement',
                        'state_topic': 'homeassistant/sensor/device_id/device_id-component_id/state',
                        'suggested_display_precision': 2,
                        'unique_id': 'device_id-component_id',
                        'unit_of_measurement': '°C',
                    },
                ),
            )
            self.assertEqual(
                sensor.get_state(),
                ComponentState(
                    topic='homeassistant/sensor/device_id/device_id-component_id/state',
                    payload=NO_STATE,  # no initial state -> NO_STATE
                ),
            )
            sensor.set_state(456)
            self.assertEqual(
                sensor.get_state(),
                ComponentState(
                    topic='homeassistant/sensor/device_id/device_id-component_id/state',
                    payload=456,  # now we have a state
                ),
            )

    def test_validation(self):
        def make_sensor(*, device_class, state_class, unit_of_measurement, validate=True):
            Sensor(
                device_class=device_class,
                state_class=state_class,
                unit_of_measurement=unit_of_measurement,
                validate=validate,
                # Needed but not relevant for this test:
                device=MqttDevice(name='My device', uid='device_id'),
                name='My component',
                uid='component_id',
            )

        with self.assertRaises(ValidationError):
            make_sensor(
                device_class='invalid',
                state_class='measurement',
                unit_of_measurement='%',
            )

        with self.assertRaises(ValidationError):
            make_sensor(
                device_class='battery',
                state_class='measurement',
                unit_of_measurement='invalid',
            )

        with self.assertRaises(ValidationError):
            make_sensor(
                device_class='data_size',
                state_class='invalid',
                unit_of_measurement='%',
            )

        make_sensor(
            device_class='invalid',
            state_class='invalid',
            unit_of_measurement='invalid',
            validate=False,
        )



import nox
from nox.sessions import Session


PYTHON_VERSIONS = ['3.13', '3.12', '3.11']


@nox.session(python=PYTHON_VERSIONS)
def tests(session: Session):
    session.install('uv')
    session.run(
        'uv',
        'sync',
        '--all-extras',
        '--python',
        session.python,
        env={'UV_PROJECT_ENVIRONMENT': session.virtualenv.location},
    )
    session.run('python', '-m', 'coverage', 'run', '--context', f'py{session.python}')


@nox.session
def lint(session: Session):
    session.install('flake8', 'flake8-bugbear')
    session.run('flake8', 'manage_django_project')



import json

from paho.mqtt.client import Client


class MqttClientMock(Client):
    def __init__(self):
        self.messages = []

    def publish(self, **kwargs) -> None:
        self.messages.append(kwargs)

    def _reset_sockets(self):
        pass

    def get_config_payload(self):
        config_payload = [
            json.loads(message['payload']) for message in self.messages if message['topic'].endswith('/config')
        ]
        return config_payload

    def get_state_messages(self):
        state_messages = [message for message in self.messages if message['topic'].endswith('/state')]
        return state_messages



from cli_base.cli_tools.code_style import assert_code_style
from cli_base.tyro_commands import TyroVerbosityArgType

from ha_services.cli_dev import PACKAGE_ROOT, app


@app.command
def lint(verbosity: TyroVerbosityArgType = 1):
    """
    Check/fix code style by run: "ruff check --fix"
    """
    assert_code_style(package_root=PACKAGE_ROOT, verbose=bool(verbosity), sys_exit=True)



import logging
from collections.abc import Callable

from paho.mqtt.client import MQTT_ERR_SUCCESS, Client, MQTTMessageInfo

from ha_services.exceptions import InvalidStateValue
from ha_services.mqtt4homeassistant.components import BaseComponent
from ha_services.mqtt4homeassistant.data_classes import NO_STATE, ComponentConfig, ComponentState
from ha_services.mqtt4homeassistant.device import MqttDevice


logger = logging.getLogger(__name__)


def default_select_callback(*, client: Client, component: BaseComponent, old_state: str, new_state: str):
    logger.info(f'{component.name} state changed: {old_state!r} -> {new_state!r}')
    component.set_state(new_state)
    component.publish_state(client)


class Select(BaseComponent):
    """
    MQTT Select component for Home Assistant.
    https://www.home-assistant.io/integrations/select.mqtt/
    """

    def __init__(
        self,
        *,
        device: MqttDevice,
        name: str,
        uid: str,
        callback: Callable = default_select_callback,
        component: str = 'select',
        initial_state=NO_STATE,  # set_state() must be called to set the value
        options: tuple[str, ...],
        default_option: str,
    ):
        super().__init__(
            device=device,
            name=name,
            uid=uid,
            component=component,
            initial_state=initial_state,
        )

        self.callback = callback

        self.options = options
        self.set_state(default_option)

        self.command_topic = f'{self.topic_prefix}/command'

    def _command_callback(self, client: Client, userdata, message: MQTTMessageInfo):
        new_state = message.payload.decode()
        assert new_state in self.options, f'Receive invalid state: {new_state!r}'

        self.callback(client=client, component=self, old_state=self.state, new_state=new_state)

    def publish_config(self, client: Client) -> MQTTMessageInfo:
        info = super().publish_config(client)

        client.message_callback_add(self.command_topic, self._command_callback)
        result, _ = client.subscribe(self.command_topic)
        if result is not MQTT_ERR_SUCCESS:
            logger.error(f'Error subscribing {self.command_topic=}: {result=}')

        return info

    def validate_state(self, state: str):
        super().validate_state(state)
        if state not in self.options:
            raise InvalidStateValue(component=self, error_msg=f'{state=} not in {self.options=}')

    def get_state(self) -> ComponentState:
        # e.g.: {'topic': 'homeassistant/select/My-device/My-Relay/state', 'payload': 'ON'}
        return ComponentState(
            topic=f'{self.topic_prefix}/state',
            payload=self.state,
        )

    def get_config(self) -> ComponentConfig:
        return ComponentConfig(
            topic=f'{self.topic_prefix}/config',
            payload={
                'component': self.component,
                'device': self.device.get_mqtt_payload(),
                'name': self.name,
                'unique_id': self.uid,
                'state_topic': f'{self.topic_prefix}/state',
                'json_attributes_topic': f'{self.topic_prefix}/attributes',
                'command_topic': self.command_topic,
                'options': self.options,
            },
        )



import dataclasses
import logging
import os
import random
import resource
import time

from cli_base.systemd.data_classes import BaseSystemdServiceInfo, BaseSystemdServiceTemplateContext
from paho.mqtt.client import Client
from rich import print  # noqa

import ha_services
from ha_services.exceptions import InvalidStateValue
from ha_services.mqtt4homeassistant.components.binary_sensor import BinarySensor
from ha_services.mqtt4homeassistant.components.select import Select
from ha_services.mqtt4homeassistant.components.sensor import Sensor
from ha_services.mqtt4homeassistant.components.switch import Switch
from ha_services.mqtt4homeassistant.data_classes import MqttSettings
from ha_services.mqtt4homeassistant.device import MainMqttDevice, MqttDevice
from ha_services.mqtt4homeassistant.mqtt import get_connected_client


logger = logging.getLogger(__name__)


@dataclasses.dataclass
class SystemdServiceTemplateContext(BaseSystemdServiceTemplateContext):
    """
    HaServices Demo - Context values for the systemd service file content.
    """

    verbose_service_name: str = 'HaServices Demo'


@dataclasses.dataclass
class SystemdServiceInfo(BaseSystemdServiceInfo):
    """
    HaServices Demo - Information for systemd helper functions.
    """

    template_context: SystemdServiceTemplateContext = dataclasses.field(default_factory=SystemdServiceTemplateContext)


@dataclasses.dataclass
class MqttExampleValues:
    """
    Some values used to create DEMO MQTT messages.
    """

    mqtt_payload_prefix: str = 'example'
    device_name: str = 'ha-services-demo'

    publish_config_throttle_seconds: int = 10  # Min. time between config publishing
    publish_throttle_seconds: int = 1  # Min. time between state publishing


@dataclasses.dataclass
class DemoSettings:
    """
    This are just settings for the "ha-services" DEMO.
    Will be used in ha_services example commands.
    See "./cli.py --help" for more information.
    """

    # Information how to setup the systemd services:
    systemd: dataclasses = dataclasses.field(default_factory=SystemdServiceInfo)

    # Information about the MQTT server:
    mqtt: dataclasses = dataclasses.field(default_factory=MqttSettings)

    # Example "app" data:
    app: dataclasses = dataclasses.field(default_factory=MqttExampleValues)


def publishing(*, user_settings: DemoSettings, verbosity: int, endless_loop: bool = True):
    """
    Publish "something" to MQTT server. It's just a DEMO ;)
    """

    main_device = MainMqttDevice(
        name='ha-services Main Device Example',
        uid=user_settings.mqtt.main_uid,
        manufacturer='ha_services',
        model='Just the example.py ;)',
        sw_version=ha_services.__version__,
        throttle_sec=user_settings.mqtt.publish_throttle_seconds,
        config_throttle_sec=user_settings.mqtt.publish_config_throttle_seconds,
    )

    device = MqttDevice(
        main_device=main_device,
        name='ha-services Sub Device Example',
        uid='ha_services_sub',
        manufacturer='ha_services',
        model='Just the example.py ;)',
        sw_version=ha_services.__version__,
        throttle_sec=user_settings.app.publish_throttle_seconds,
        config_throttle_sec=user_settings.app.publish_config_throttle_seconds,
    )

    activate_relay = BinarySensor(
        device=device,
        name='Activate Relay',
        uid='activate_relay',
        device_class=None,  # None == Generic ON/OFF or e.g.: 'door', 'motion', etc...
    )

    relay_delay = Sensor(
        device=device,
        name='Relay "delay"',
        uid='relay_delay',
        state_class='measurement',
        unit_of_measurement='sec',
    )

    select = Select(
        device=device,
        name='Select',
        uid='select',
        options=('Option 1', 'Option 2', 'Option 3'),
        default_option='Option 1',
    )

    def relay_callback(*, client: Client, component: Switch, old_state: str, new_state: str):
        logger.info(f'{component.name} state changed: {old_state!r} -> {new_state!r}')
        delay = random.randrange(5)
        logger.info(f'{delay=}')

        if not activate_relay.is_on:
            logger.info('Relay is not activated!')
            return

        relay_delay.set_state(delay)
        relay_delay.publish(mqttc)
        time.sleep(delay)

        component.set_state(new_state)
        component.publish_state(client)

    relay = Switch(
        device=device,
        name='Virtual Relay',
        uid='relay',
        callback=relay_callback,
    )
    system_load_sensor = Sensor(
        device=device,
        name='System load 1min.',
        uid='system_load',
        state_class='measurement',
        suggested_display_precision=2,
    )
    user_time_used_sensor = Sensor(
        device=device,
        name='Time in user mode (float seconds)',
        uid='user_time_used',
        state_class='measurement',
        unit_of_measurement='sec',
        suggested_display_precision=2,
    )
    system_time_used_sensor = Sensor(
        device=device,
        name='Time in system mode (float seconds)',
        uid='system_time_used',
        state_class='measurement',
        unit_of_measurement='sec',
        suggested_display_precision=2,
    )

    mqttc = get_connected_client(settings=user_settings.mqtt, verbosity=verbosity)
    mqttc.loop_start()

    relay.set_state(relay.OFF if random.randrange(2) else relay.ON)

    while True:
        try:
            main_device.poll_and_publish(mqttc)

            activate_relay.set_state(relay.OFF if random.randrange(2) else relay.ON)
            activate_relay.publish(mqttc)

            relay.publish(mqttc)
            select.publish(mqttc)

            system_load_sensor.set_state(os.getloadavg()[0])
            system_load_sensor.publish(mqttc)

            usage = resource.getrusage(resource.RUSAGE_SELF)
            user_time_used_sensor.set_state(usage.ru_utime)
            user_time_used_sensor.publish(mqttc)

            system_time_used_sensor.set_state(usage.ru_stime)
            system_time_used_sensor.publish(mqttc)
        except InvalidStateValue as err:
            logger.warning('Skip invalid state: %s', err)
        else:
            if not endless_loop:
                logger.info('Exiting...')
                break

        print('\n', flush=True)
        print('Wait', end='...')
        for i in range(10, 1, -1):
            time.sleep(0.5)
            print(i, end='...', flush=True)

    return main_device



import logging

from cli_base.cli_tools.verbosity import setup_logging
from cli_base.tyro_commands import TyroVerbosityArgType
from rich import print  # noqa

from ha_services.cli_app import app
from ha_services.mqtt4homeassistant.system_info.wifi_info import get_wifi_infos


logger = logging.getLogger(__name__)


@app.command
def wifi_info(verbosity: TyroVerbosityArgType):
    """
    Just display the WiFi info
    """
    setup_logging(verbosity=verbosity)

    wifi_infos = get_wifi_infos(verbosity=verbosity)
    print(wifi_infos)



import logging
from unittest import TestCase

from ha_services.exceptions import InvalidStateValue
from ha_services.mqtt4homeassistant.components.switch import Switch
from ha_services.mqtt4homeassistant.data_classes import NO_STATE, ComponentConfig, ComponentState
from ha_services.mqtt4homeassistant.device import MqttDevice


class SwitchTestCase(TestCase):
    def test_switch(self):
        with self.assertNoLogs(level=logging.WARNING):
            switch = Switch(
                device=MqttDevice(name='My Device', uid='device1'),
                name='My Switch',
                uid='switch1',
            )
            self.assertEqual(
                switch.get_config(),
                ComponentConfig(
                    topic='homeassistant/switch/device1/device1-switch1/config',
                    payload={
                        'command_topic': 'homeassistant/switch/device1/device1-switch1/command',
                        'component': 'switch',
                        'device': {'identifiers': 'device1', 'name': 'My Device'},
                        'json_attributes_topic': 'homeassistant/switch/device1/device1-switch1/attributes',
                        'name': 'My Switch',
                        'payload_off': 'OFF',
                        'payload_on': 'ON',
                        'state_topic': 'homeassistant/switch/device1/device1-switch1/state',
                        'unique_id': 'device1-switch1',
                    },
                ),
            )
            self.assertEqual(
                switch.get_state(),
                ComponentState(
                    topic='homeassistant/switch/device1/device1-switch1/state',
                    payload=NO_STATE,  # no initial state -> NO_STATE
                ),
            )
            with self.assertRaises(InvalidStateValue) as cm:
                switch.set_state('invalid_state')
            self.assertEqual(
                str(cm.exception), "Switch(self.uid='device1-switch1'): state='invalid_state' not in ON, OFF"
            )

            self.assertEqual(
                switch.get_state(),
                ComponentState(
                    topic='homeassistant/switch/device1/device1-switch1/state',
                    payload=NO_STATE,  # still NO_STATE
                ),
            )

            switch.set_state('ON')
            self.assertEqual(
                switch.get_state(),
                ComponentState(
                    topic='homeassistant/switch/device1/device1-switch1/state',
                    payload='ON',  # now we have a state
                ),
            )



import logging
from unittest import TestCase
from unittest.mock import patch

from bx_py_utils.test_utils.snapshot import assert_snapshot

from ha_services.mqtt4homeassistant.device import MqttDevice
from ha_services.mqtt4homeassistant.mocks import HostSystemMock
from ha_services.mqtt4homeassistant.mocks.mqtt_client_mock import MqttClientMock
from ha_services.mqtt4homeassistant.system_info import wifi_info
from ha_services.mqtt4homeassistant.system_info.wifi_info import (
    WifiInfo,
    WifiInfo2Mqtt,
    WifiInfoValue,
    _convert_iwconfig_values,
    _get_iwconfig_values,
    get_wifi_infos,
)


class WifiInfoTestCase(TestCase):
    def test_happy_path(self):
        with HostSystemMock(), self.assertNoLogs(level=logging.WARNING):
            iwconfig_values = _get_iwconfig_values()
        self.assertEqual(
            iwconfig_values,
            {
                'wlo1': {
                    'ESSID': 'foobar',
                    'bit_rate': '29.2',
                    'bit_rate_unit': 'Mb/s',
                    'frequency': '5.18',
                    'frequency_unit': 'GHz',
                    'link_quality': '45/70',
                    'signal_level': '-65',
                    'signal_level_unit': 'dBm',
                }
            },
        )

        ##########################################################################################

        WIFI_INFOS = [
            WifiInfo(
                device_name='wlo1',
                values=[
                    WifiInfoValue(slug='ESSID', name='Essid', value='foobar', unit=None),
                    WifiInfoValue(slug='bit_rate', name='Bit rate', value=29.2, unit='Mb/s'),
                    WifiInfoValue(slug='frequency', name='Frequency', value=5.18, unit='GHz'),
                    WifiInfoValue(slug='link_quality', name='Link quality', value='45/70', unit=None),
                    WifiInfoValue(slug='signal_level', name='Signal level', value=-65, unit='dBm'),
                ],
            )
        ]

        self.assertEqual(_convert_iwconfig_values(iwconfig_values), WIFI_INFOS)

        ##########################################################################################

        with HostSystemMock(), self.assertNoLogs(level=logging.WARNING):
            wifi_infos = get_wifi_infos()
        self.assertEqual(wifi_infos, WIFI_INFOS)

        ##########################################################################################

        with self.assertNoLogs(level=logging.WARNING):
            mqtt_client_mock = MqttClientMock()
            wifi_info2mqtt = WifiInfo2Mqtt(
                device=MqttDevice(name='Main Device', uid='main_uid'),
            )

            with patch.object(wifi_info, 'get_wifi_infos', return_value=WIFI_INFOS):
                wifi_info2mqtt.poll_and_publish(client=mqtt_client_mock)

        # Some pre-checks:
        topics = [msg['topic'] for msg in mqtt_client_mock.messages]
        self.assertEqual(
            topics,
            [
                'homeassistant/sensor/main_uid/main_uid-wifi_device_name/config',
                'homeassistant/sensor/main_uid/main_uid-wifi_device_name/state',
                'homeassistant/sensor/main_uid/main_uid-ESSID/config',
                'homeassistant/sensor/main_uid/main_uid-ESSID/state',
                'homeassistant/sensor/main_uid/main_uid-bit_rate/config',
                'homeassistant/sensor/main_uid/main_uid-bit_rate/state',
                'homeassistant/sensor/main_uid/main_uid-frequency/config',
                'homeassistant/sensor/main_uid/main_uid-frequency/state',
                'homeassistant/sensor/main_uid/main_uid-link_quality/config',
                'homeassistant/sensor/main_uid/main_uid-link_quality/state',
                'homeassistant/sensor/main_uid/main_uid-signal_level/config',
                'homeassistant/sensor/main_uid/main_uid-signal_level/state',
            ],
        )
        for message in mqtt_client_mock.messages:
            topic = message['topic']
            payload = message['payload']
            self.assertIsInstance(payload, (int, float, str), message)
            if topic.endswith('/state'):
                if 'bit_rate' in topic:
                    self.assertEqual(payload, 29.2)
                elif 'frequency' in topic:
                    self.assertEqual(payload, 5.18)

        assert_snapshot(got=mqtt_client_mock.messages)



import json
import logging
from unittest import TestCase

import frozendict
from bx_py_utils.test_utils.snapshot import assert_snapshot
from freezegun.api import freeze_time

from ha_services.mqtt4homeassistant.components.sensor import Sensor
from ha_services.mqtt4homeassistant.data_classes import NO_STATE, ComponentConfig, ComponentState
from ha_services.mqtt4homeassistant.device import MainMqttDevice, MqttDevice
from ha_services.mqtt4homeassistant.mocks import HostSystemMock
from ha_services.mqtt4homeassistant.mocks.mqtt_client_mock import MqttClientMock


class IntergrationTestCase(TestCase):

    def test_main_sub(self):
        with (
            self.assertNoLogs(level=logging.WARNING),
            HostSystemMock(),
            freeze_time(
                time_to_freeze='2012-01-14T12:00:00+00:00',
                tick=True,  # Needed to avoid ZeroDivisionError in rate calculation
            ),
        ):
            main_device = MainMqttDevice(
                name='Main Device',
                uid='main_uid',
            )
            device = MqttDevice(
                main_device=main_device,
                name='Sub Device',
                uid='sub_uid',
            )
            test_sensor = Sensor(
                device=device,
                name='Test Sensor',
                uid='sensor_uid',
            )
            self.assertEqual(
                test_sensor.get_state(),
                ComponentState(
                    topic='homeassistant/sensor/main_uid-sub_uid/main_uid-sub_uid-sensor_uid/state',
                    payload=NO_STATE,  # No initial state -> NO_STATE
                ),
            )
            test_sensor.set_state(123)
            self.assertEqual(
                test_sensor.get_state(),
                ComponentState(
                    topic='homeassistant/sensor/main_uid-sub_uid/main_uid-sub_uid-sensor_uid/state',
                    payload=123,
                ),
            )

            self.assertEqual(
                test_sensor.get_config(),
                ComponentConfig(
                    topic='homeassistant/sensor/main_uid-sub_uid/main_uid-sub_uid-sensor_uid/config',
                    payload={
                        'component': 'sensor',
                        'device': frozendict.frozendict(
                            {
                                'name': 'Sub Device',
                                'identifiers': 'main_uid-sub_uid',
                                'via_device': 'main_uid',
                            }
                        ),
                        'device_class': None,
                        'json_attributes_topic': (
                            'homeassistant/sensor/main_uid-sub_uid/main_uid-sub_uid-sensor_uid/attributes'
                        ),
                        'name': 'Test Sensor',
                        'state_class': None,
                        'state_topic': 'homeassistant/sensor/main_uid-sub_uid/main_uid-sub_uid-sensor_uid/state',
                        'unique_id': 'main_uid-sub_uid-sensor_uid',
                        'unit_of_measurement': None,
                    },
                ),
            )

            mqtt_client_mock = MqttClientMock()

            # throttle values not set, yet:
            self.assertEqual(main_device.hostname._next_config_publish, 0)
            self.assertEqual(main_device.hostname._next_publish, 0)

            main_device.poll_and_publish(mqtt_client_mock)

            # Now the throttle values should be set:
            self.assertGreater(main_device.hostname._next_config_publish, 0)
            self.assertGreater(main_device.hostname._next_publish, 0)

        first_message = mqtt_client_mock.messages[0]
        self.assertEqual(first_message['topic'], 'homeassistant/sensor/main_uid/main_uid-hostname/config')
        self.assertEqual(
            json.loads(first_message['payload']),
            {
                'component': 'sensor',
                'device': {'identifiers': 'main_uid', 'name': 'Main Device'},
                'device_class': None,
                'json_attributes_topic': 'homeassistant/sensor/main_uid/main_uid-hostname/attributes',
                'name': 'Hostname',
                'origin': {
                    'name': 'ha-services-tests',
                    'support_url': 'https://pypi.org/project/ha_services/',
                    'sw_version': '1.2.3',
                },
                'state_class': None,
                'state_topic': 'homeassistant/sensor/main_uid/main_uid-hostname/state',
                'unique_id': 'main_uid-hostname',
                'unit_of_measurement': None,
            },
        )

        for message in mqtt_client_mock.messages:
            topic = message['topic']
            payload = message['payload']
            self.assertIsInstance(payload, (int, float, str), message)

            if topic.endswith('/state'):
                if 'eth0' in topic:
                    message['payload'] = '<mocked eth0 values>'
                elif 'up_time' in topic:
                    assert payload.startswith('2009-02-13T23:'), f'{payload=}'
                    message['payload'] = '2009-02-13T23:<mocked-ticks>'

        assert_snapshot(got=mqtt_client_mock.messages)
